"""LGPL library for small arms ballistic calculations (Python 3.8+)"""

__author__ = "o-murphy"
__copyright__ = ("",)

__credits__ = ["o-murphy"]
__version__ = "1.1.3"


from .backend import *
from .drag_tables import *
from .settings import *
from .multiple_bc import *
from .interface import *
from .trajectory_data import *
from .conditions import *
from .munition import *
from .unit import *
